// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr


struct data {
size_t n;
double * y;
double * sigma;
};

int expb_f (const gsl_vector * x, void *data, gsl_vector * f)
{
size_t n = ((struct data *)data)->n;
double *y = ((struct data *)data)->y;
double *sigma = ((struct data *) data)->sigma;

double A = gsl_vector_get (x, 0);
double lambda = gsl_vector_get (x, 1);
double b = gsl_vector_get (x, 2);

size_t i;

for (i = 0; i < n; i++)
{
	/* Model Yi = A * exp(-lambda * i) + b */
	double t = i;
	double Yi = A * exp (-lambda * t) + b;
	gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);

}

return GSL_SUCCESS;
}

int expb_df (const gsl_vector * x, void *data, gsl_matrix * J)
{
size_t n = ((struct data *)data)->n;
double *sigma = ((struct data *) data)->sigma;

double A = gsl_vector_get (x, 0);
double lambda = gsl_vector_get (x, 1);

size_t i;

for (i = 0; i < n; i++)
{
	/* Jacobian matrix J(i,j) = dfi / dxj, */
	/* where fi = (Yi - yi)/sigma[i],      */
	/*       Yi = A * exp(-lambda * i) + b  */
	/* and the xj are the parameters (A,lambda,b) */
	double t = i;
	double s = sigma[i];
	double e = exp(-lambda * t);
	gsl_matrix_set (J, i, 0, e/s);
	gsl_matrix_set (J, i, 1, -t * A * e/s);
	gsl_matrix_set (J, i, 2, 1/s);
printf("J = %d\n",i);
}
return GSL_SUCCESS;
}

int expb_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J)
{
expb_f (x, data, f);
expb_df (x, data, J);

return GSL_SUCCESS;
}

#define N 40

void print_state (size_t iter, gsl_multifit_fdfsolver * s)
{
printf ("iter: %3u x = % 15.8f % 15.8f % 15.8f "
       "|f(x)| = %g\n",
       iter,
       gsl_vector_get (s->x, 0),
       gsl_vector_get (s->x, 1),
       gsl_vector_get (s->x, 2),
       gsl_blas_dnrm2 (s->f));
}


struct istruct main_g;
struct istruct main_exp;
struct istruct main_error;
struct istruct main_fit;
static double t0=4e-6;
static double v0=1e5;
static double sigma=1e-20;

int fft_len=1024*2*2*2*2*2;
//Replaces data[1..2*nn] by its discrete Fourier transform, if isign is input as 1; or replaces
//data[1..2*nn] by nn times its inverse discrete Fourier transform, if isign is input as −1.
//data is a complex array of length nn or, equivalently, a real array of length 2*nn. nn MUST
//be an integer power of 2 (this is not checked for!).
static double error_tot=0.0;

double sech(double x)
{
return 2.0/(exp(x)+exp(-x));
}

double get_fft_error (const gsl_vector *v, void *params)
{
int n;
for (n=0;n<main_g.len;n++)
{
	main_g.data[n]=fabs(gsl_vector_get(v, n));
}
t0=fabs(gsl_vector_get(v, n));
int i;
int ii;


double En=0.0;
double Ep=0.0;
double E=0.0;
double dE=0.0;
double Eomega=0.0;
double sum=0.0;

for (ii=1;ii<main_exp.len;ii++)
{

Eomega=main_exp.x[ii];
sum=0.0;
	for (i=1;i<main_g.len;i++)
	{

		if (i==1)
		{
			En=main_g.x[i];
		}else
		{
			En=main_g.x[i-1];
		}

		if (i==main_g.len-1)
		{
			Ep=main_g.x[i];
		}else
		{
			Ep=main_g.x[i+1];
		}

		E=main_g.x[i];
		dE=fabs((Ep-En)/2.0);

		sum+=(sigma*v0/2.0)*main_g.data[i]*sech(q*(Eomega-E)/(300.0*kb))*dE;
	}

main_fit.x[ii]=main_exp.x[ii];
main_fit.data[ii]=sum;

main_error.x[ii]=main_exp.x[ii];
main_error.data[ii]=fabs(main_exp.data[ii]-main_fit.data[ii]);
}
error_tot=0.0;
for (i=0;i<main_exp.len;i++)
{
if ((main_error.x[i]>0.0)&&(main_error.x[i]<0.14))
{
error_tot+=main_error.data[i];
}
}

inter_save_a(&main_exp,"./","rod_exp.dat");
inter_save_a(&main_fit,"./","rod_fit.dat");
inter_save_a(&main_error,"./","rod_error.dat");
//printf("Dump\n");
//getchar();
return error_tot;
}

double fft_mag(struct istruct* in,int pos)
{
return sqrt(in->data[pos*2]*in->data[pos*2]+in->data[(pos*2)+1]*in->data[(pos*2)+1]);
}



void fft_extract()
{
const gsl_multifit_fdfsolver_type *T;
gsl_multifit_fdfsolver *s;
int status;
unsigned int i, iter = 0;
const size_t n = N;
const size_t p = 3;

gsl_matrix *covar = gsl_matrix_alloc (p, p);
double y[N], sigma[N];
struct data d = { n, y, sigma};
gsl_multifit_function_fdf f;
double x_init[3] = { 1.0, 0.0, 0.0 };
gsl_vector_view x = gsl_vector_view_array (x_init, p);
const gsl_rng_type * type;
gsl_rng * r;

gsl_rng_env_setup();

type = gsl_rng_default;
r = gsl_rng_alloc (type);

f.f = &expb_f;
f.df = &expb_df;
f.fdf = &expb_fdf;
f.n = n;
f.p = p;
f.params = &d;

/* This is the data to be fitted */

for (i = 0; i < n; i++)
 {
   double t = i;
   y[i] = 1.0 + 5 * exp (-0.1 * t)
              + gsl_ran_gaussian (r, 0.1);
   sigma[i] = 0.1;
   printf ("data: %u %g %g\n", i, y[i], sigma[i]);
 };

T = gsl_multifit_fdfsolver_lmsder;
s = gsl_multifit_fdfsolver_alloc (T, n, p);
gsl_multifit_fdfsolver_set (s, &f, &x.vector);

print_state (iter, s);

do
 {
   iter++;
   status = gsl_multifit_fdfsolver_iterate (s);

   printf ("status = %s\n", gsl_strerror (status));

   print_state (iter, s);

   if (status)
     break;

   status = gsl_multifit_test_delta (s->dx, s->x,
                                     1e-4, 1e-4);
 }
while (status == GSL_CONTINUE && iter < 500);

gsl_multifit_covar (s->J, 0.0, covar);

#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))

{
 double chi = gsl_blas_dnrm2(s->f);
 double dof = n - p;
 double c = GSL_MAX_DBL(1, chi / sqrt(dof));

 printf("chisq/dof = %g\n",  pow(chi, 2.0) / dof);

 printf ("A      = %.5f +/- %.5f\n", FIT(0), c*ERR(0));
 printf ("lambda = %.5f +/- %.5f\n", FIT(1), c*ERR(1));
 printf ("b      = %.5f +/- %.5f\n", FIT(2), c*ERR(2));
}

printf ("status = %s\n", gsl_strerror (status));

gsl_multifit_fdfsolver_free (s);
gsl_matrix_free (covar);
gsl_rng_free (r);
return 0;
}



void fft_new(struct istruct* in)
{
inter_new(in,fft_len*2);
}

void fft_to_mag(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	//printf("%le\n",in->data[i*2]);
	//getchar();
	out->data[i]=fft_mag(in,i);

	out->x[i]=in->x[i*2];
}


}

void fft_to_phase(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	out->data[i]=atan2(in->data[(i*2)+1],in->data[(i*2)]);
	out->x[i]=in->x[i*2];
}


}

void fft_to_real(struct istruct* in,struct istruct* out)
{

int i;
for (i=0;i<in->len/2;i++)
{
	out->data[i]=in->data[i*2];
	out->x[i]=in->x[i*2];
}


}

void i_to_fft(struct istruct* out,struct istruct* in)
{
int i=0;
for (i=0;i<in->len;i++)
{
	out->data[i*2]=in->data[i];
	out->x[i*2]=in->x[i];
	out->x[(i*2)+1]=in->x[i];
}

}

void do_dft(struct istruct* out,struct istruct* in)
{
int i;
int ii;
double tn=0.0;
double tp=0.0;
double dt=0.0;
double t=0.0;
double omega=0.0;
double a=0.0;
double b=0.0;

double delta=(in->x[((in->len/2-1)*2)]-in->x[(0*2)])/(in->len/2);//in->x[(1*2)]-in->x[(0*2)];//
for (i=0;i<out->len/2;i++)
{
out->data[i*2]=0.0;
out->data[(i*2)+1]=0.0;
omega=2.0*pi*((double)i/(double)(out->len/2))/delta;
//printf("%le",omega);
//getchar();
//if (i==0) omega=0.0;
	for (ii=0;ii<in->len/2;ii++)
	{
		if (ii==0)
		{
			tn=0.0;
		}else
		{
			tn=in->x[(ii*2)-1];
		}

		if (ii==in->len-1)
		{
			tp=in->x[(ii*2)]+delta/2.0;
		}else
		{
			tp=in->x[(ii*2)+1];
		}

		dt=(tp-tn)/2.0;
		t=in->x[ii*2];

		a=in->data[ii*2];
		b=in->data[(ii*2)+1];
		//out->data[i*2]+=(a*cos(omega*t)+b*sin(omega*t))*dt;
		//out->data[(i*2)+1]+=(-a*sin(omega*t)+b*cos(omega*t))*dt;
		out->data[i*2]+=(a*cos(omega*t)-b*sin(omega*t))*dt;
		out->data[(i*2)+1]+=(a*sin(omega*t)+b*cos(omega*t))*dt;
		//printf("%le %d %d\n",in->x[i*2],i,ii);
		//getchar();


	}

	out->x[i*2]=omega;
	out->x[(i*2)+1]=omega;

		//printf("%le %le\n",omega,sqrt(out->data[i*2]*out->data[i*2]+out->data[(i*2)+1]*out->data[(i*2)+1]));
		//getchar();
}

}

void do_fft(struct istruct* in, int isign)
{
unsigned long nn=in->len/2;
double *data=in->data-1;
unsigned long n,mmax,m,j,istep,i;
double wtemp,wr,wpr,wpi,wi,theta;
//Double precision for the trigonomet-ric recurrences.
double tempr,tempi;
n=nn << 1;
j=1;
for (i=1;i<n;i+=2)
{
	//This is the bit-reversal section of the
		if (j > i)
		{
			//routine.
			SWAP(data[j],data[i]);
			//Exchange the two complex numbers.
			SWAP(data[j+1],data[i+1]);
		}

		m=n >> 1;
		while (m >= 2 && j > m)
		{
			j -= m;
			m >>= 1;
		}
	j += m;
}
//Here begins the Danielson-Lanczos section of the routine.
mmax=2;
while (n > mmax)
{
	//Outer loop executed log2 nn times.
	istep=mmax << 1;
	theta=isign*(6.28318530717959/mmax);
	//Initialize the trigonometric recurrence.
	wtemp=sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi=sin(theta);
	wr=1.0;
	wi=0.0;
	for (m=1;m<mmax;m+=2)
	{
			for (i=m;i<=n;i+=istep)
			{
				j=i+mmax;
				tempr=wr*data[j]-wi*data[j+1];
				tempi=wr*data[j+1]+wi*data[j];
				data[j]=data[i]-tempr;
				data[j+1]=data[i+1]-tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	mmax=istep;
	//Here are the two nested inner loops. This is the Danielson-Lanczos formula: Trigonometric recurrence.
}

double delta=(in->x[(1*2)]-in->x[(0*2)]);
printf("%le %le %le\n",delta,in->x[(1*2)],in->x[(0*2)]);
if (isign==-1)
{
	for (i=1;i<in->len;i++)
	{
	in->data[i]/=(double)fft_len;
	}

}else
{

	for (i=0;i<in->len/2;i++)
	{
		in->x[(i*2)]=2.0*pi*((double)i)/((double)(in->len/2)*delta);
		in->x[(i*2)+1]=2.0*pi*((double)i)/((double)(in->len/2)*delta);
		//printf("%le\n",delta*(double)i);
	}
}

}

