// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "sim.h"

double get_mat_param(struct layer* in,int param)
{
if (in->mat[param].len==0)
{
printf("No data!\n");
exit(0);
}
//printf("%d\n",in->mat[param].len);
return in->mat[param].c[0];
}

double get_mat_param_poly(struct layer* in,int param,double x)
{
int i;
double mysum=0.0;

for (i=0;i<in->mat[param].len;i++)
{
mysum+=in->mat[param].c[i]*pow(x,i);
//printf("sum=%e\n",in->mat[param].c[i]*pow(x,i));
//getchar();
}

//printf("end sum %e\n",mysum);
//getchar();

return mysum;
}

double get_mat_param_xy(struct layer* in,int param,double x)
{
double mysum=0.0;

if (in->mat[param].len==2)
{
	//printf("%e %e %e %e\n",in->mat[param].c[0]*pow(x,in->mat[param].c[1]),in->mat[param].c[1],in->mat[param].c[0],x,in->mat[param].c[1]);
	//getchar();

	return in->mat[param].c[0]*pow(x,in->mat[param].c[1]);
}else
{
printf("error\n");
exit(0);
}

return mysum;
}

double get_mat_param_mobility(struct layer* in,int param,int n)
{
if (in->mat[param].len!=3)
{
	printf("Not enough bands!\n");
	exit(0);
}

double ret=in->mat[param].c[0]+(in->mat[param].c[1]-in->mat[param].c[0])*((double)n)/((double)in->mat[param].c[2]);

return ret;
}

double get_mat_param_dmobility(struct layer* in,int param,double n)
{
if (in->mat[param].len==1)
{
	return 0.0;
}else
if (in->mat[param].len==2)
{
return in->mat[param].c[1]*in->mat[param].c[0]*pow(n,in->mat[param].c[1]-1.0);
}else
{
printf("error\n");
exit(0);
}
return 0;
}

double get_mat_param_lin(struct layer* in,int param,double x)
{
double ret=0.0;
if (in->mat[param].len!=5)
{
printf("For get_mat_param_lin I need 5 data elements!\n");
exit(0);
}
//format is as so
//c0 x0 x1 y0 y2
double c0=in->mat[param].c[0];
double x0=in->mat[param].c[1];
double x1=in->mat[param].c[2];
double y0=in->mat[param].c[3];
double y1=in->mat[param].c[4];
ret=c0+((y1-y0)/(x1-x0))*(x-x0);

return ret;
}


void mat_free(struct layer* in)
{
int i;
	for (i=0;i<100;i++)
	{
//printf("%d\n",in->mat[i].used);
		if (in->mat[i].used==TRUE)
		{
			free(in->mat[i].c);
			in->mat[i].used=FALSE;
			in->mat[i].c=NULL;
			in->mat[i].len=0;
		}
	}
}

void mat_load_poly(FILE *file,struct param* in)
{
if (in->used==TRUE)
{
printf("space full\n");
exit(0);
}
in->used=TRUE;
char buffer[1000];
fgets(buffer, 1000, file);
fscanf(file,"%d",&(in->len));
//printf("segment of length= %d %s\n",in->len,buffer);
in->c = malloc(in->len*sizeof(double));
int i;
for (i=0;i<in->len;i++)
{
	fscanf(file,"%le",&(in->c[i]));
}

}

void trimstring(char * in)
{
int i;
for (i=0;i<strlen(in);i++)
{
if (in[i]==10) in[i]=0;
}
}

void load_mat_file(char *name,struct layer* in)
{
FILE *file;
if (!(file=fopen(name,"r")))
{
	printf("Error opening file %s\n",name);
	exit(0);
}
char buffer[1000];

int i;
for (i=0;i<100;i++)
{
in->mat[i].len=0;
in->mat[i].c=NULL;
in->mat[i].used=FALSE;
}

do
{
	fgets(buffer, 1000, file);
	trimstring(buffer);
	//printf("-%s-\n",buffer);
	//getchar();
	if (strcmp(buffer,"[Eg]")==0)
	{
		//printf("found Eg\n");
		mat_load_poly(file,&(in->mat[mat_Eg]));
	}else
	if (strcmp(buffer,"[Xi]")==0)
	{
		//printf("found Xi\n");
		mat_load_poly(file,&(in->mat[mat_Xi]));
	}else
	if (strcmp(buffer,"[mue]")==0)
	{
		//printf("found mue\n");
		mat_load_poly(file,&(in->mat[mat_mue]));
	}else
	if (strcmp(buffer,"[muh]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_muh]));
	}else
	if (strcmp(buffer,"[me]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_me]));
	}else
	if (strcmp(buffer,"[mh]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_mh]));
	}else
	if (strcmp(buffer,"[epsilonr]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_epsilonr]));
	}else
	if (strcmp(buffer,"[Dex]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_Dex]));
	}else
	if (strcmp(buffer,"[kf]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_kf]));
	}else
	if (strcmp(buffer,"[a]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_a]));
	}else
	if (strcmp(buffer,"[kl]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_kl]));
	}else
	if (strcmp(buffer,"[taue0]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_taue0]));
	}else
	if (strcmp(buffer,"[taue1]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_taue1]));
	}else
	if (strcmp(buffer,"[taup0]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_tauh0]));
	}else
	if (strcmp(buffer,"[taup1]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_tauh1]));
	}
	else
	if (strcmp(buffer,"[Xiphoton]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_Egphoton]));
	}else
	if (strcmp(buffer,"[Egphoton]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_Xiphoton]));
	}	else
	if (strcmp(buffer,"[A]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_A]));
	}else
	if (strcmp(buffer,"[deltaE]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_deltaE]));
	}else
	if (strcmp(buffer,"[capeff]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_capeff]));
	}else
	if (strcmp(buffer,"[br]")==0)
	{
		//printf("found mun\n");
		mat_load_poly(file,&(in->mat[mat_br]));
	}else
	if (strcmp(buffer,"[dostype]")==0)
	{
		mat_load_poly(file,&(in->mat[mat_dostype]));
	}else
	if (strcmp(buffer,"[Rbi]")==0)
	{
		mat_load_poly(file,&(in->mat[mat_rbi]));
	}


}while(!feof(file));

fclose(file);
}
