// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

static double glob_wanted=0.0;
static int glob_use_cap=0;
static int ntricks_fast;

void set_ntricks_fast(int val)
{
ntricks_fast=val;
}

void ramp(struct device *in,double from,double to,double steps)
{
in->Vapplied=from;
double dV=(to-from)/steps;
if (dV>1.0) dV=1.0;
printf("dV=%le\n",dV);
printf("Ramping: from=%e to=%e\n",from,to);
if (load_state(in,to)==TRUE)
{
solve_all(in);
save_state(in,to);
return;
}

if (fabs(from-to)<0.2) return;

do
{
in->Vapplied+=dV;
printf("ramp: %lf\n",in->Vapplied);
solve_all(in);

//sim_externalv(in,in->cevoltage);

if (fabs(in->Vapplied-to)<0.15)
{
save_state(in,to);
return;
}
}while(1);


}


void save_state(struct device *in,double to)
{
printf("Dumping state\n");
int i;
int band;
FILE *state;
state=fopena(in->outputpath,"./state.dat","w");

fprintf(state,"%le ",to);


for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(state,"%le %le %le ",in->phi[i],in->x[i],in->xp[i]);

	for (band=0;band<in->srh_bands;band++)
	{
		fprintf(state,"%le %le ",in->xt[i][band],in->xpt[i][band]);
	}

}
fclose(state);
}

int load_state(struct device *in,double voltage)
{
printf("Load state\n");
int i;
int band;
double vtest;
FILE *state;
state=fopena(in->outputpath,"./state.dat","r");
if (!state)
{
printf("State not found\n");
return FALSE;
}

fscanf(state,"%le",&vtest);
printf("%le %le",voltage,vtest);
if (vtest!=voltage)
{
printf("State not found\n");
return FALSE;
}
printf("Loading state\n");

in->Vapplied=vtest;

for (i=0;i<in->ymeshpoints;i++)
{
	fscanf(state,"%le %le %le ",&(in->phi[i]),&(in->x[i]),&(in->xp[i]));

	for (band=0;band<in->srh_bands;band++)
	{
		fscanf(state,"%le %le ",&(in->xt[i][band]),&(in->xpt[i][band]));
	}

}
fclose(state);
return TRUE;
}


//    ----=----
//    |       |-----
//    |       |    |
//    |       |   ---
//    |       SC
//    |       |   ---
//    |       |    |
//  Vtot      |-----
//    |       R
//    |	      |
//    ----=----
void newton_aux_celiv(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double C=in->C;
if (glob_use_cap==FALSE) C=0.0;
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=glob_wanted-(V+((in->Rcontact+in->Rshort_celiv)*(i0+V/in->Rshunt+C*(V-in->Vapplied_last)/in->dt)));
*didv=-(1.0+((in->Rcontact+in->Rshort_celiv)*(didv0+1.0/in->Rshunt+C*(1.0)/in->dt)));
*didphi=-((in->Rcontact+in->Rshort_celiv)*didphi0);
*didxil=-((in->Rcontact+in->Rshort_celiv)*didxil0);
*didxipl=-((in->Rcontact+in->Rshort_celiv)*didxipl0);
*didphir=-((in->Rcontact+in->Rshort_celiv)*didphir0);
*didxir=-((in->Rcontact+in->Rshort_celiv)*didxir0);
*didxipr=-((in->Rcontact+in->Rshort_celiv)*didxipr0);
return;
}


double newton_celiv(struct device *in,double Vtot,int usecap)
{
double C=in->C;
in->newton_cur=TRUE;
solver_realloc(in);
glob_wanted=Vtot;
glob_use_cap=usecap;
in->newton_aux=&newton_aux_celiv;
solve_all(in);
if (glob_use_cap==FALSE) C=0.0;
return get_I(in)+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt;
}

//    ----  ----
//    |        |
//    |        |
//   SC        R
//    |        |
//    |	       |
//    ----  ----
//So we hand the solver everyting as if we were going to do current and it hads it back as it wasnts.
void newton_aux_ce(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double C=in->C;
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;

double S=((in->Rshort+in->Rcontact)-in->Rshunt)/(in->Rshunt*(in->Rshort+in->Rcontact));

*i=i0+C*(V-in->Vapplied_last)/in->dt-V*S;
*didv=didv0+C*(1.0)/in->dt-S;
//*i=i0-V*S;
//*didv=didv0-S;
*didphi=didphi0;
*didxil=didxil0;
*didxipl=didxipl0;

return;
}

double newton_charge_extraction(struct device *in)
{
in->newton_cur=TRUE;
solver_realloc(in);
in->newton_aux=&newton_aux_ce;
solve_all(in);
return 0.0;
}

double newton_sim_tpv_wire(struct device *in)
{
double step=0.001;
double e0;
double e1;
double i0;
double i1;
double deriv;
solve_all(in);
i0=get_I(in);
//getchar();
e0=fabs(i0+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
deriv=(e1-e0)/step;
step=-e1/deriv;
in->Vapplied+=step;
int count=0;
int max=100;
do
{
e0=e1;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
deriv=(e1-e0)/step;
step=-e1/deriv;
//if (e1<clamp) clamp=e1/10.0;
//step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
printf("Tpv error = %e %e\n",in->Vapplied,e1);
if (count>max) break;
count++;
}while(e1>1e-8);
double cur=wire_sim(in->Vapplied);
printf("result = %e %e %e %e\n",i1,in->Vapplied/in->Rshunt,cur,in->Vapplied/cur);
//getchar();
return 0.0;
}

void newton_aux_i(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=glob_wanted-i0-V/in->Rshunt;
*didv=-didv0-1.0/in->Rshunt;
*didphi=-didphi0;
*didxil=-didxil0;
*didxipl=-didxipl0;
*didphir=-didphir0;
*didxir=-didxir0;
*didxipr=-didxipr0;
return;
}


double sim_i(struct device *in,double wantedi)
{
in->newton_cur=TRUE;
solver_realloc(in);
glob_wanted=wantedi;
in->newton_aux=&newton_aux_i;
solve_all(in);
return 0.0;
}

void newton_aux_voc(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=V/in->Rshunt-i0;
*didv=1.0/in->Rshunt-didv0;
*didphi=-didphi0;
*didxil=-didxil0;
*didxipl=-didxipl0;
*didphir=-didphir0;
*didxir=-didxir0;
*didxipr=-didxipr0;
return;
}

double sim_voc(struct device *in)
{
in->newton_cur=TRUE;
solver_realloc(in);
in->newton_aux=&newton_aux_voc;
solve_all(in);
printf("Voc=%le\n",get_equiv_V(in));
return 0.0;
}

void newton_aux_externalv(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;

*i=glob_wanted-(V+in->Rcontact*(i0+V/in->Rshunt));
*didv=-(1.0+in->Rcontact*(didv0+1.0/in->Rshunt));
*didphi=-(in->Rcontact*(didphi0));
*didxil=-(in->Rcontact*(didxil0));
*didxipl=-(in->Rcontact*(didxipl0));
*didphir=-(in->Rcontact*(didphir0));
*didxir=-(in->Rcontact*(didxir0));
*didxipr=-(in->Rcontact*(didxipr0));
return;
}

double sim_externalv(struct device *in,double wantedv)
{
in->newton_cur=TRUE;
solver_realloc(in);

glob_wanted=wantedv;
in->newton_aux=&newton_aux_externalv;
solve_all(in);
return 0.0;
}

void newton_aux_tpv(struct device *in,double V,double* i,double* didv,double* didphi,double* didxil,double* didxipl,double* didphir,double* didxir,double* didxipr)
{
double i0=*i;
double didv0=*didv;
double didphi0=*didphi;
double didxil0=*didxil;
double didxipl0=*didxipl;
double didphir0=*didphir;
double didxir0=*didxir;
double didxipr0=*didxipr;
double Rdrain=1.0/(1.0/in->Rshunt+1.0/1e6);

*i=i0-V/Rdrain;
*didv=didv0-1.0/Rdrain;
*didphi=didphi0;
*didxil=didxil0;
*didxipl=didxipl0;
*didphir=didphir0;
*didxir=didxir0;
*didxipr=didxipr0;
return;
}

void newton_sim_tpv_fast(struct device *in)
{
in->newton_cur=TRUE;
solver_realloc(in);

in->newton_aux=&newton_aux_tpv;
solve_all(in);
return 0.0;
}

double newton_sim_tpv(struct device *in)
{
double C=in->C;
double clamp=0.1;
double step=0.05;
double e0;
double e1;
double i0;
double i1;
double deriv;
double Rdrain=1.0/(1.0/in->Rshunt+1.0/1e6);
solve_all(in);
i0=get_I(in);
e0=fabs(i0-in->Vapplied/Rdrain);

in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
e1=fabs(i1-in->Vapplied/Rdrain);

deriv=(e1-e0)/step;
step=-e1/deriv;

step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;

int count=0;
int max=1000;
do
{
e0=e1;
solve_all(in);
i1=get_I(in);
e1=fabs(i1-in->Vapplied/Rdrain);
deriv=(e1-e0)/step;
step=-e1/deriv;

step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
//printf("%d %le\n",count,e1);
if (count>max) break;
count++;
}while(e1>1e-12);

double ret=in->Vapplied-C*(i1-in->Ilast)/in->dt;
return ret;
}

//</clean>
